theory Graphtransfo_rel imports Routing

begin

(* ---------- Morphisms, Transformations, Rules ---------- *)

type_synonym 'obj graphmorph = "nat \<Rightarrow> 'obj option"

definition
  morph_sep :: "['obj graphmorph, nat set] \<Rightarrow> bool" where
  "morph_sep gm S = (\<forall> x y. x \<in> S \<longrightarrow> y \<notin> S \<longrightarrow> gm x \<noteq> gm y)"

(* ---------- Structure-modifying transformations ---------- *)

record ('nt, 'et) graphtrans = 
  -- "applicability condition"
  appcond :: "('nt, 'et) path_form"
  -- "mapping of nodes"
  ndel  :: "nat set"   -- "deleted nodes"
  ngen  :: "nat set"   -- "generated nodes"
  -- "mapping of edges"
  edel  :: "(nat * nat) set"
  egen  :: "(nat * nat) set"

definition
  graphtrans_gen_nodes :: 
  "[('nt, 'et, 'a) graphtrans_scheme, 
  ('obj, 'b) graph_scheme, 'obj graphmorph, 'obj set] \<Rightarrow> bool" where
  "graphtrans_gen_nodes gt gr morph_gen gen_nodes =
    ((dom morph_gen = ngen gt) \<and> (ran morph_gen = gen_nodes) \<and> 
    (gen_nodes \<inter> nodes gr = {}))"

definition 
  graphtrans_gen_nodes_sd:: "[nat set, 'obj set, 'obj set, 'obj graphmorph] \<Rightarrow> bool" where
  "graphtrans_gen_nodes_sd nds_dom nds_ran nds_gr morph_gen  =
    ((dom morph_gen = nds_dom) \<and> (nds_ran = ran morph_gen) \<and> (nds_ran \<inter> nds_gr = {}))"


lemma graphtrans_gen_nodes_init: "graphtrans_gen_nodes gt gr morph_gen gen_nodes = 
  graphtrans_gen_nodes_sd (ngen gt) gen_nodes (nodes gr) morph_gen"
by (simp add: graphtrans_gen_nodes_def graphtrans_gen_nodes_sd_def) fast


lemma graphtrans_gen_nodes_sd_empty:
  "graphtrans_gen_nodes_sd {} nds_ran nds_gr morph_gen = (morph_gen = empty \<and> nds_ran = {})"
by (simp add: graphtrans_gen_nodes_sd_def cong: conj_cong)

lemma graphtrans_gen_nodes_sd_insert: "a \<notin> A \<Longrightarrow>
  graphtrans_gen_nodes_sd (insert a A) nds_ran nds_gr morph_gen = 
  (\<exists>b gm' nds_ran'. morph_gen = gm'(a \<mapsto> b) \<and> nds_ran = nds_ran' \<union> {b} \<and> b \<notin> nds_gr \<and>
  graphtrans_gen_nodes_sd A nds_ran' nds_gr gm')"
apply (simp add: graphtrans_gen_nodes_sd_def)
apply (simp add: dom_reduce_insert)
apply (simp add: ex_simps [THEN sym] del: ex_simps cong: conj_cong)
apply (simp add: ran_map_upd_notin_dom cong: conj_cong)
apply fast
done

definition
  apply_graphtrans_rel_nodes_for_gen :: 
  "[('nt, 'et, 'a) graphtrans_scheme, 'obj graphmorph, 
            ('obj, 'b) graph_scheme, ('obj, 'b) graph_scheme,
            'obj graphmorph, 'obj set] \<Rightarrow> bool" where
  "apply_graphtrans_rel_nodes_for_gen gt gm gr gr' morph_gen gen_nodes =
    ((graphtrans_gen_nodes gt gr morph_gen gen_nodes) \<and>
    (let del_nodes = ran (gm |` (ndel gt)) in
    let nds = ((nodes gr) - del_nodes) \<union> gen_nodes in 
    let del_edges = ((emorph gm) ` (edel gt)) in
    let gen_edges = ((emorph (gm ++ morph_gen)) ` (egen gt)) in

  (nodes gr' = nds) \<and>
  (edges gr' = (restrict_rel (edges gr) nds - del_edges) \<union> gen_edges)
  ))"

definition
  -- "removes dangling edges (restrict_rel)"
  apply_graphtrans_rel_utpd :: "[('nt, 'et, 'a) graphtrans_scheme, 'obj graphmorph, 'obj graphmorph \<Rightarrow> bool,
            ('obj, 'b) graph_scheme, ('obj, 'b) graph_scheme] \<Rightarrow> bool" where
  "apply_graphtrans_rel_utpd gt gm morph_prop gr gr' =
  (\<exists> morph_gen gen_nodes.
  morph_prop (gm ++ morph_gen) \<and>
  apply_graphtrans_rel_nodes_for_gen gt gm gr gr' morph_gen gen_nodes)"


(* ---------- Properties of transformations ---------- *)

definition
  -- "structurally well-formed graph transformation"
  struct_wf_gt_utpd :: "('nt, 'et, 'a) graphtrans_scheme \<Rightarrow> bool" where
  "struct_wf_gt_utpd gt == 
    (ndel gt) \<subseteq> (fv_path_form (appcond gt)) \<and>
    finite (ngen gt) \<and> (fv_path_form (appcond gt)) \<inter> (ngen gt) = {} \<and>
    (egen gt) \<inter> (edel gt) = {} \<and>
    (Field (edel gt)  \<subseteq>  (fv_path_form (appcond gt))) \<and>
    (Field (egen gt)  \<subseteq>  ((fv_path_form (appcond gt)) - (ndel gt)) \<union> (ngen gt))"


(* ---------- Properties of morphisms ---------- *)


(* Well-formedness *)

   (* TODO: proposal for locale extensions fixing type variables
locale graph_transformation = struct_wf_graph gr for gr :: "('obj, 'a) graph_scheme"  +
  fixes I_nt :: "'nt \<Rightarrow> 'obj set"
  fixes I_et :: "'et \<Rightarrow> ('obj * 'obj) set"
 *)

locale graph_transformation_loc = 
  graph_interpretation_loc gr I_nt I_et 
  for gr :: "('obj, 'a) graph_scheme"
  and I_nt :: "'nt \<Rightarrow> 'obj set"
  and I_et :: "'et \<Rightarrow> ('obj * 'obj) set" +
  fixes gt :: "('nt, 'et, 'b) graphtrans_scheme"
  fixes apply_graphtrans_rel :: "['obj graphmorph,  ('obj, 'a) graph_scheme] \<Rightarrow> bool"
  fixes morph_prop :: "'obj graphmorph \<Rightarrow> bool"

context graph_transformation_loc
begin

definition 
  applicable_gm :: "['obj graphmorph, ('nt, 'et) path_form] \<Rightarrow> bool" where
  -- "graph morphism gm is applicable for path_form pf in graph gr"
  "applicable_gm gm pf =
     ((dom gm = fv_path_form pf) \<and> (ran gm \<subseteq> nodes gr) \<and> 
       morph_prop gm \<and>
       path_form_interp (the o gm) pf)"


definition
  applicable_transfo :: bool where
  "applicable_transfo  = (\<exists> gm. applicable_gm gm (appcond gt))"

definition
  apply_transfo_rel  :: "('obj, 'a) graph_scheme \<Rightarrow> bool" where
  "apply_transfo_rel  gr' = apply_graphtrans_rel (SOME gm. applicable_gm gm (appcond gt)) gr'"


 -- "For getting rid of the choice operator"
lemma apply_applicable_transfo_rel:
  "\<lbrakk> applicable_transfo \<and> apply_transfo_rel gr';
        \<And>gm. \<lbrakk> applicable_gm gm (appcond gt); apply_graphtrans_rel gm gr'\<rbrakk>  \<Longrightarrow> Q gr' \<rbrakk> 
  \<Longrightarrow>  Q gr'"
apply (unfold apply_transfo_rel_def applicable_transfo_def)
apply (erule conjE)
apply (drule someI_ex)
apply fast
done


(* -----------  Simplification of applicability condition  -------------- *)

definition
  "applicable_gm_red gm' A gm pf P == 
  (P \<and> (dom gm' = A) \<and> path_form_interp (the o gm) pf)"


(* --- Simplification of graphtrans_gen_nodes ------- *)

lemma applicable_gm_red_init: "applicable_gm gm pf = 
  (applicable_gm_red gm (fv_path_form pf) gm pf 
   ((ran gm \<subseteq> nodes gr) \<and> morph_prop gm))"
by (fastforce simp add: applicable_gm_def applicable_gm_red_def)


lemma applicable_gm_red_insert: "applicable_gm_red gm' (insert a A) gm pf P = 
  (\<exists> b gm''. applicable_gm_red gm'' A gm pf ((gm' = gm''(a\<mapsto>b)) \<and> P))"
apply (simp add: applicable_gm_red_def dom_reduce_insert)
apply (fastforce simp add: ex_simps [THEN sym] del: ex_simps cong: conj_cong)
done

lemma applicable_gm_red_empty:
  "applicable_gm_red gm' {} gm pf P = 
  (gm' = empty \<and> P \<and> path_form_interp (the o gm) pf)"
by (fastforce simp add: applicable_gm_red_def)

lemmas applicable_gm_red_simps =
  applicable_gm_red_init applicable_gm_red_insert applicable_gm_red_empty

end -- "context graph_transformation_loc"


locale graph_transformation_swfgrgt_loc = graph_transformation_loc +
  assumes struct_wf_gr: "struct_wf_gr_utpd gr"
  assumes struct_wf_gt: "struct_wf_gt_utpd gt"
  defines apply_graphtrans_rel: 
  "apply_graphtrans_rel gm gr' == apply_graphtrans_rel_utpd gt gm morph_prop gr gr'"

sublocale graph_transformation_swfgrgt_loc \<subseteq> graph_swfgr_loc
using struct_wf_gr by unfold_locales 

context graph_transformation_swfgrgt_loc
begin
(* Validity of graphs is preserved by application of graph transformation *)


lemma struct_wf_gt_pres_finite_rel: 
  "\<lbrakk> apply_graphtrans_rel gm gr'\<rbrakk>  \<Longrightarrow> finite (nodes gr')"
using struct_wf_gr struct_wf_gt
apply (simp add: apply_graphtrans_rel_def apply_graphtrans_rel_nodes_for_gen_def
  apply_graphtrans_rel_utpd_def struct_wf_gt_utpd_def graphtrans_gen_nodes_def Let_def)
apply (fastforce intro: finite_dom_ran struct_wf_gr_finite_nodes)
done

lemma struct_wf_gt_struct_wf_gr_rel_aux:
  "\<lbrakk>
  graphtrans_gen_nodes gt gr morph_gen gen_nodes; 
  applicable_gm gm (appcond gt); morph_sep gm (ndel gt)
 \<rbrakk>
  \<Longrightarrow> Field (emorph (gm ++ morph_gen) ` egen gt)
          \<subseteq> nodes gr - ran (gm |` ndel gt) \<union> gen_nodes"
using struct_wf_gt

apply (simp add: Field_emorph)
apply (simp add: struct_wf_gt_utpd_def)
apply (simp add: applicable_gm_def)

apply (subgoal_tac "(the \<circ>  (gm ++ morph_gen)) ` Field (egen gt) \<subseteq>
   (the \<circ>  (gm ++ morph_gen)) ` (fv_path_form (appcond gt) - ndel gt \<union> ngen gt)")
  prefer 2 apply (simp add:  image_mono) 
apply (erule subset_trans)
apply (simp add: image_Un)

apply (clarsimp simp add: graphtrans_gen_nodes_def)

apply (simp add: map_add_image_ran)
apply (subgoal_tac "(the \<circ> gm ++ morph_gen) ` (fv_path_form (appcond gt) - ndel gt) = (the \<circ> gm) ` (fv_path_form (appcond gt) - ndel gt)")
  prefer 2 apply (rule the_map_add_image) apply simp apply fast


apply (clarsimp)

apply (rule conjI)
apply (subgoal_tac "xa \<in> dom gm") prefer 2 apply simp apply (subgoal_tac "the (gm xa) \<in> ran gm") apply fast apply (simp (no_asm_simp) add: ran_def) apply fast

apply (simp add: ran_restrict_iff)
apply (simp add: morph_sep_def)
done

theorem struct_wf_gt_struct_wf_gr_rel: 
  assumes  "applicable_gm gm (appcond gt)"
  and "morph_sep gm (ndel gt)"
  and "apply_graphtrans_rel gm gr'"
  shows "struct_wf_gr_utpd gr'"
using assms struct_wf_gr struct_wf_gt
apply (simp (no_asm_simp) add: struct_wf_gr_utpd_def)
apply (rule conjI)
apply (erule struct_wf_gt_pres_finite_rel)


apply (simp add: apply_graphtrans_rel_def apply_graphtrans_rel_nodes_for_gen_def
  apply_graphtrans_rel_utpd_def Let_def)
apply (elim conjE exE)
apply (simp add: Field_Un)
apply (intro allI conjI)
apply (rule subset_trans) apply (rule Field_Diff_subseteq) 
apply simp
apply (rule struct_wf_gt_struct_wf_gr_rel_aux) 
apply fast apply assumption+
done

end -- "context graph_transformation_swfgrgt_loc"


    (* "untyped" here refers to the instantiation of node and edge types, not to struct_wf_gr_utpd  *)
locale untyped_graph_transformation_loc = 
  graph_transformation_loc _ "(\<lambda> _. UNIV)" "(\<lambda> _. UNIV)" _ _ _ +
  defines apply_graphtrans_rel: 
  "apply_graphtrans_rel gm gr' == apply_graphtrans_rel_utpd gt gm morph_prop gr gr'"

sublocale untyped_graph_transformation_loc \<subseteq> untyped_graph_interpretation_loc
done

locale untyped_graph_transformation_swfgrgt_loc = untyped_graph_transformation_loc +
  assumes struct_wf_gr: "struct_wf_gr_utpd gr"
  assumes struct_wf_gt: "struct_wf_gt_utpd gt"

  (* Proved more generally above
context untyped_graph_transformation_swfgrgt_loc
begin


lemma struct_wf_gt_pres_finite_rel: 
  "\<lbrakk> apply_graphtrans_rel gm gr'\<rbrakk>  \<Longrightarrow> finite (nodes gr')"
using struct_wf_gr struct_wf_gt
apply (simp add: apply_graphtrans_rel_def apply_graphtrans_rel_nodes_for_gen_def
  apply_graphtrans_rel_utpd_def struct_wf_gt_utpd_def graphtrans_gen_nodes_def Let_def)
apply (fastforce intro: finite_dom_ran struct_wf_gr_finite_nodes)
done

lemma struct_wf_gt_struct_wf_gr_rel_aux:
  "\<lbrakk>
  graphtrans_gen_nodes gt gr morph_gen gen_nodes; 
  applicable_gm gm (appcond gt); morph_sep gm (ndel gt)
 \<rbrakk>
  \<Longrightarrow> Field (emorph (gm ++ morph_gen) ` egen gt)
          \<subseteq> nodes gr - ran (gm |` ndel gt) \<union> gen_nodes"
using struct_wf_gt

apply (simp add: Field_emorph)
apply (simp add: struct_wf_gt_utpd_def)
apply (simp add: applicable_gm_def)

apply (subgoal_tac "(the \<circ>  (gm ++ morph_gen)) ` Field (egen gt) \<subseteq>
   (the \<circ>  (gm ++ morph_gen)) ` (fv_path_form (appcond gt) - ndel gt \<union> ngen gt)")
  prefer 2 apply (simp add:  image_mono) 
apply (erule subset_trans)
apply (simp add: image_Un)

apply (clarsimp simp add: graphtrans_gen_nodes_def)

apply (simp add: map_add_image_ran)
apply (subgoal_tac "(the \<circ> gm ++ morph_gen) ` (fv_path_form (appcond gt) - ndel gt) = (the \<circ> gm) ` (fv_path_form (appcond gt) - ndel gt)")
  prefer 2 apply (rule the_map_add_image) apply simp apply fast


apply (clarsimp)

apply (rule conjI)
apply (subgoal_tac "xa \<in> dom gm") prefer 2 apply simp apply (subgoal_tac "the (gm xa) \<in> ran gm") apply fast apply (simp (no_asm_simp) add: ran_def) apply fast

apply (simp add: ran_restrict_iff)
apply (simp add: morph_sep_def)
done

theorem struct_wf_gt_struct_wf_gr_rel:
  assumes  "applicable_gm gm (appcond gt)"
  and "morph_sep gm (ndel gt)"
  and "apply_graphtrans_rel gm gr'"
  shows "struct_wf_gr_utpd gr'"
using assms struct_wf_gr struct_wf_gt
apply (simp (no_asm_simp) add: struct_wf_gr_utpd_def)
apply (rule conjI)
apply (erule struct_wf_gt_pres_finite_rel)


apply (simp add: apply_graphtrans_rel_def apply_graphtrans_rel_nodes_for_gen_def
  apply_graphtrans_rel_utpd_def Let_def)
apply (elim conjE exE)
apply (simp add: Field_Un)
apply (intro allI conjI)
apply (rule subset_trans) apply (rule Field_Diff_subseteq) 
apply simp
apply (rule struct_wf_gt_struct_wf_gr_rel_aux) 
apply fast apply assumption+
done

end -- "context untyped_graph_transformation_swfgrgt_loc"
 *)

end
