(* Routing expressions in graphs - locale version *)

theory Routing imports Graphs

begin

datatype 'nt nodeset
  = Univ_set              -- "set of all nodes"                 ("\<lceil>Univs\<rceil>")
  | Singleton_set nat    -- "singleton containing constant"     ("\<guillemotleft> _ \<guillemotright>"  [60] 60)
  | Node_type_set 'nt    -- "set of all nodes of given type"    ("\<lceil>\<Colon> _ \<rceil>"  [60] 60)

(* Paths and path formulas *)
datatype 'et path 
  = Ident_pth                          -- "identity path"            ("\<lceil>Idp\<rceil>")
  | Univ_pth                           -- "path between all nodes"   ("\<lceil>Univp\<rceil>")
  | Edge_pth nat nat                   -- "path between two nodes"   ("\<guillemotleft>/ _, _/\<guillemotright>"  [60,60] 60)
  | Edge_type_pth 'et                  -- "edge with given edge type"  ("\<lceil>\<Colon> _ \<Colon>\<rceil>"  [60] 60)
  | Compl_pth  "('et path)"            -- "complement path"           ("\<lceil>-\<rceil> _"  [40] 40)
  | Conv_pth  "('et path)"             -- "converse path"             ("(_\<lceil>\<inverse>\<rceil>)" [1000] 999)
  | Seq_pth "('et path)" "('et path)"  -- "sequential composition"    (infixr "\<lceil>;\<rceil>" 35)
  | Alt_pth "('et path)" "('et path)"  -- "alternative"               (infixr "\<lceil>+\<rceil>" 35)
  | Clos_pth "('et path)"              -- "transitive closure"        ("(_\<lceil>^+\<rceil>)" [1000] 999)


datatype ('nt, 'et) path_form 
  = S_form nat "('nt nodeset)"                                  ("(_/ \<lceil>\<in>\<rceil> _)" [50, 51] 50)
  | P_form nat nat "('et path)"                                 ("(_/ \<leadsto>  _/ \<lceil>\<in>\<rceil> _)" [50, 50, 51] 50)
  | Neg_form "(('nt, 'et) path_form)"                           ("\<lceil>\<not>\<rceil> _"  [40] 40)
  | Conj_form "(('nt, 'et) path_form)" "(('nt, 'et) path_form)" (infixr "\<lceil>\<and>\<rceil>" 35)
  | All_form "(('nt, 'et) path_form)"                           ("\<lceil>\<forall>\<rceil> _"  [10] 10)


definition
  -- "constant true"
  trueF :: "('nt, 'et) path_form" ("\<lceil>\<top>\<rceil>") where
  "trueF = All_form (S_form 0 (Singleton_set 0))"
definition
  -- "disjunction"
  Disj_form :: "[('nt, 'et) path_form,  ('nt, 'et) path_form] \<Rightarrow> ('nt, 'et) path_form" (infixr "\<lceil>\<or>\<rceil>" 35) where
  "Disj_form pf pf' = Neg_form (Conj_form (Neg_form pf) (Neg_form pf'))"
definition
  -- "implication"
  Imp_form :: "[('nt, 'et) path_form,  ('nt, 'et) path_form]  \<Rightarrow> ('nt, 'et) path_form" (infixr "\<lceil>\<longrightarrow>\<rceil>" 35) where
  "Imp_form pf pf' = Disj_form (Neg_form pf) pf'"
definition
  -- "existential quantifier"
  Ex_form :: "('nt, 'et) path_form \<Rightarrow> ('nt, 'et) path_form" ("\<lceil>\<exists>\<rceil> _"  [10] 10) where
  "Ex_form pf = (Neg_form (All_form (Neg_form pf)))"


  (* Some abbreviations and syntax *)
abbreviation is_node ("\<lceil>\<guillemotleft> _ \<guillemotright>\<rceil>"  [60] 60)
  where "\<lceil>\<guillemotleft>n\<guillemotright>\<rceil> == (S_form n (Singleton_set n))"
abbreviation is_edge ("\<lceil>\<guillemotleft>/ _, _/ \<guillemotright>\<rceil>"  [60,60] 60)
  where "\<lceil>\<guillemotleft>n1, n2\<guillemotright>\<rceil> == (P_form n1 n2 (Edge_pth n1 n2))"

syntax
  -- {* list Enumeration *}
  "_path_form" :: "args => ('nt, 'et) path_form"    ("\<lceil>{_}\<rceil>")

translations
  "\<lceil>{n, ns }\<rceil>" == "\<lceil>\<guillemotleft> n \<guillemotright>\<rceil> \<lceil>\<and>\<rceil> \<lceil>{ns}\<rceil>"
  "\<lceil>{n}\<rceil>" == "\<lceil>\<guillemotleft> n \<guillemotright>\<rceil>"


  -- "node n has type t"
abbreviation
  nd_tpF :: "[nat, 'nt] \<Rightarrow> ('nt, 'et) path_form" ("\<lceil> _  \<Colon> _\<rceil>"  [40, 40] 40) where
  "(\<lceil>n \<Colon> t\<rceil>) == S_form n (Node_type_set t)"

  -- "edge n1, n2 has type t"
abbreviation
  edge_tpF :: "[nat, nat, 'et] \<Rightarrow> ('nt, 'et) path_form" ("\<lceil>\<guillemotleft>/ _, _/ \<guillemotright> \<Colon> _\<rceil>"  [40, 40, 40] 40) where
  "(\<lceil>\<guillemotleft>n1, n2\<guillemotright> \<Colon> t\<rceil>) == (P_form n1 n2 (Edge_type_pth t))"

  -- "edge typing of e"
definition
  edge_relF :: "'et \<Rightarrow> 'nt \<Rightarrow> 'nt \<Rightarrow> ('nt, 'et) path_form" ("\<lceil> _ \<subseteq> _ \<Rightarrow> _\<rceil>"  [40, 40, 40] 40) where
  "(\<lceil> e \<subseteq> t \<Rightarrow> t'\<rceil>) = (\<lceil>\<forall>\<rceil> \<lceil>\<forall>\<rceil> (\<lceil>\<guillemotleft>0, 1\<guillemotright> \<Colon> e\<rceil>) \<lceil>\<longrightarrow>\<rceil> (\<lceil>0 \<Colon> t\<rceil>) \<lceil>\<and>\<rceil> (\<lceil>1 \<Colon> t'\<rceil>))"


(* free variables *)
fun fv_nodeset :: "'nt nodeset \<Rightarrow> nat set" where
  "fv_nodeset (Singleton_set n) = {n}"
| "fv_nodeset _ = {}"

fun fv_path :: "'et path \<Rightarrow> nat set" where
  "fv_path Ident_pth = {}"
| "fv_path Univ_pth = {}"
| "fv_path (Edge_pth n1 n2) = {n1, n2}"
| "fv_path (Edge_type_pth e) = {}"
| "fv_path (Compl_pth p) = fv_path p"
| "fv_path (Conv_pth p) = fv_path p"
| "fv_path (Seq_pth p1 p2) = (fv_path p1) \<union> (fv_path p2)"
| "fv_path (Alt_pth p1 p2) = (fv_path p1) \<union> (fv_path p2)"
| "fv_path (Clos_pth p) = fv_path p"

fun fv_path_form :: "('nt, 'et) path_form \<Rightarrow> nat set" where
  "fv_path_form (S_form n s) = {n} \<union> fv_nodeset s"
| "fv_path_form (P_form n n' p) = {n, n'} \<union> fv_path p"
| "fv_path_form (Neg_form pf) = fv_path_form pf"
| "fv_path_form (Conj_form pf pf') = fv_path_form pf \<union> fv_path_form pf'"
| "fv_path_form (All_form pf) = (\<lambda> x. x - 1) ` ((fv_path_form pf) - {0})"

lemma [simp]: "fv_path_form trueF = {}"
by (simp add: trueF_def)
lemma [simp]: "fv_path_form (Disj_form pf pf') = fv_path_form pf \<union> fv_path_form pf'"
by (simp add: Disj_form_def)
lemma [simp]: "fv_path_form (Imp_form pf pf') = fv_path_form pf \<union> fv_path_form pf'"
by (simp add: Imp_form_def)
lemma [simp]: "fv_path_form (Ex_form pf) = (\<lambda> x. x - 1) ` ((fv_path_form pf) - {0})"
by (simp add: Ex_form_def)

lemma finite_fv_nodeset [simp, intro]: "finite (fv_nodeset s)"
by (induct s) simp_all

lemma finite_fv_path [simp, intro]: "finite (fv_path p)"
by (induct p) simp_all

lemma finite_fv_path_form [simp, intro]: "finite (fv_path_form pf)"
by (induct pf) simp_all


(* ______________________________________________________________________ *)
(* Interpretations of Paths and path formulas *)


locale graph_interpretation_loc = 
  fixes gr :: "('obj, 'a) graph_scheme"
  fixes I_nt :: "'nt \<Rightarrow> 'obj set"
  fixes I_et :: "'et \<Rightarrow> ('obj * 'obj) set"

context graph_interpretation_loc
begin 

fun nodeset_interp :: "[nat \<Rightarrow> 'obj,  'nt nodeset] \<Rightarrow> 'obj set" where
  "nodeset_interp I \<lceil>Univs\<rceil> = nodes gr"
| "nodeset_interp I (\<guillemotleft> n \<guillemotright>) = (nodes gr \<inter> {I n})"
| "nodeset_interp I (\<lceil>\<Colon> nt\<rceil>) = (nodes gr \<inter> I_nt nt)"


fun path_interp :: "[nat \<Rightarrow> 'obj, 'et path] \<Rightarrow> ('obj * 'obj) set" where
  "path_interp I \<lceil>Idp\<rceil> = Id_on (nodes gr)"
| "path_interp I \<lceil>Univp\<rceil> = (edges gr)"
| "path_interp I (\<guillemotleft>n1, n2\<guillemotright>) = (edges gr \<inter> {(I n1, I n2)})"
| "path_interp I (\<lceil>\<Colon> et \<Colon>\<rceil>) = (edges gr \<inter> I_et et)"
| "path_interp I (\<lceil>-\<rceil> p) = ((nodes gr) \<times> (nodes gr)) - (path_interp I p)"
| "path_interp I (p\<lceil>\<inverse>\<rceil>) = (path_interp I p)^-1"
| "path_interp I (p \<lceil>;\<rceil> p') = (path_interp I p) O (path_interp I p')"
| "path_interp I (p \<lceil>+\<rceil> p') = (path_interp I p) \<union> (path_interp I p')"
| "path_interp I (p\<lceil>^+\<rceil>) = (path_interp I p)^+"

fun path_form_interp :: "[nat \<Rightarrow> 'obj, ('nt, 'et) path_form] \<Rightarrow> bool" where
  "path_form_interp I (n \<lceil>\<in>\<rceil> s) = (I n \<in> nodeset_interp I s)"
| "path_form_interp I (n \<leadsto> n' \<lceil>\<in>\<rceil> p) = ((I n, I n') \<in> path_interp I p)"
| "path_form_interp I (\<lceil>\<not>\<rceil> pf) = (\<not> (path_form_interp I pf))"
| "path_form_interp I (pf \<lceil>\<and>\<rceil> pf') = ((path_form_interp I pf) \<and> (path_form_interp I pf'))"
| "path_form_interp I (\<lceil>\<forall>\<rceil> pf) = 
       (\<forall> x. x \<in> nodes gr \<longrightarrow> path_form_interp ((I \<circ> (\<lambda> x. x - 1))(0:=x)) pf)"

lemma [simp]: "path_form_interp I trueF = True"
by (simp add: trueF_def Id_on_def)

lemma [simp]: "path_form_interp I (Disj_form pf pf') = 
        ((path_form_interp I pf) \<or> (path_form_interp I pf'))"
by (simp add: Disj_form_def)

lemma [simp]: "path_form_interp I (Imp_form pf pf') = 
        ((path_form_interp I pf) \<longrightarrow> (path_form_interp I pf'))"
by (simp add: Imp_form_def)

lemma [simp]: "path_form_interp I (Ex_form pf) = 
   (\<exists> x. x \<in> nodes gr \<and> path_form_interp ((I \<circ> (\<lambda> x. x - 1))(0:=x)) pf)"
by (simp add: Ex_form_def)

end -- "context graph_interpretation_loc"

locale untyped_graph_interpretation_loc
  = graph_interpretation_loc _  "(\<lambda> _. UNIV)" "(\<lambda> _. UNIV)"


end
